@@ -5,26 +5,17 @@ __all__ = ['affinegap',
            'clustering',
            'core',
            'lr',
+           'backport',
            'predicates',
            'crossvalidation',
-           'distance'
+           'distance',
+           'centroid'
            ]
 
 #from distance import affinegap
-from distance import affinegap
-import distance 
-import blocking
-import clustering
-import core
-import lr
-import predicates
-import training
-import crossvalidation
-import datamodel
-import backport
 from api import StaticDedupe, Dedupe
 from api import StaticRecordLink, RecordLink
+from api import StaticGazetteer, Gazetteer
 from core import randomPairs
-from convenience import consoleLabel, trainingDataDedupe, trainingDataLink
+from convenience import consoleLabel, trainingDataDedupe, trainingDataLink, canonicalize
 from AsciiDammit import asciiDammit
-import backport
