@@ -13,9 +13,12 @@ import itertools
 import logging
 import pickle
 import numpy
+import multiprocessing
 import random
 import warnings
 import copy
+import os
+
 try:
     from collections import OrderedDict
 except ImportError :
@@ -30,6 +33,7 @@ import dedupe.predicates as predicates
 import dedupe.blocking as blocking
 import dedupe.clustering as clustering
 from dedupe.datamodel import DataModel
+import dedupe.centroid as centroid
 
 logger = logging.getLogger(__name__)
 
@@ -48,6 +52,12 @@ class Matching(object):
         self.matches = None
         self.blocker = None
 
+    def __del__(self) :
+        try :
+            os.remove(self.matches.filename)
+        except :
+            pass
+
     def thresholdBlocks(self, blocks, recall_weight=1.5):
         """
         Returns the threshold that maximizes the expected F score,
@@ -68,7 +78,7 @@ class Matching(object):
 
         probability = core.scoreDuplicates(self._blockedPairs(blocks), 
                                            self.data_model, 
-                                           self.num_processes)['score']
+                                           self.num_cores)['score']
 
         probability.sort()
         probability = probability[::-1]
@@ -89,7 +99,7 @@ class Matching(object):
 
         return probability[i]
 
-    def matchBlocks(self, blocks, threshold=.5):
+    def matchBlocks(self, blocks, threshold=.5, *args, **kwargs):
         """
         Partitions blocked data and returns a list of clusters, where
         each cluster is a tuple of record ids
@@ -117,43 +127,23 @@ class Matching(object):
         
         self.matches = core.scoreDuplicates(candidate_records,
                                             self.data_model,
-                                            self.num_processes,
+                                            self.num_cores,
                                             threshold)
 
-        clusters = self._cluster(self.matches, cluster_threshold)
+        logger.info("matching done, begin clustering")
+
+        clusters = self._cluster(self.matches, 
+                                 cluster_threshold, *args, **kwargs)
         
         return clusters
 
     def _checkRecordType(self, record) :
-        for k in self.data_model.field_comparators :
-            if k not in record :
+        for field, _ in self.data_model.field_comparators :
+            if field not in record :
                 raise ValueError("Records do not line up with data model. "
                                  "The field '%s' is in data_model but not "
-                                 "in a record" % k)
+                                 "in a record" % field)
 
-    def _blockedPairs(self, blocks) :
-        """
-        Generate tuples of pairs of records from a block of records
-        
-        Arguments:
-        
-        blocks -- an iterable sequence of blocked records
-        """
-        
-        block, blocks = core.peek(blocks)
-        self._checkBlock(block)
-
-        def pair_gen() :
-            disjoint = set.isdisjoint
-            blockPairs = self._blockPairs
-            for block in blocks :
-                for pair in blockPairs(block) :
-                    ((key_1, record_1, smaller_ids_1), 
-                     (key_2, record_2, smaller_ids_2)) = pair
-                    if disjoint(smaller_ids_1, smaller_ids_2) :
-                        yield (key_1, record_1), (key_2, record_2)
-
-        return pair_gen()
 
     def _logLearnedWeights(self): # pragma: no cover
         """
@@ -162,9 +152,9 @@ class Matching(object):
         logger.info('Learned Weights')
         for (key_1, value_1) in self.data_model.items():
             try:
-                for (key_2, value_2) in value_1.items():
-                    logger.info((key_2, value_2.weight))
-            except AttributeError:
+                for field in value_1 :
+                    logger.info((field.name, field.weight))
+            except TypeError:
                 logger.info((key_1, value_1))
 
 
@@ -191,8 +181,10 @@ class DedupeMatching(Matching) :
     def match(self, data, threshold = 0.5) : # pragma : no cover
         """
         Identifies records that all refer to the same entity, returns tuples
-        of record ids, where the record_ids within each tuple should refer
-        to the same entity
+        containing a set of record ids and a confidence score as a float between 0
+        and 1. The record_ids within each set should refer to the
+        same entity and the confidence score is a measure of our confidence that
+        all the records in a cluster refer to the same entity.
         
         This method should only used for small to moderately sized datasets
         for larger data, use matchBlocks
@@ -234,9 +226,24 @@ class DedupeMatching(Matching) :
         blocked_pairs = self._blockData(data)
         return self.thresholdBlocks(blocked_pairs, recall_weight)
 
-    def _blockPairs(self, block) :  # pragma : no cover
-        return itertools.combinations(block, 2)
+    def _blockedPairs(self, blocks) :
+        """
+        Generate tuples of pairs of records from a block of records
         
+        Arguments:
+        
+        blocks -- an iterable sequence of blocked records
+        """
+        
+        block, blocks = core.peek(blocks)
+        self._checkBlock(block)
+
+	combinations = itertools.combinations
+
+        pairs = (combinations(block, 2) for block in blocks)
+
+        return itertools.chain.from_iterable(pairs) 
+
     def _checkBlock(self, block) :
         if not block :
             raise ValueError("You have not provided any data blocks")
@@ -264,24 +271,41 @@ class DedupeMatching(Matching) :
         blocks = OrderedDict({})
         coverage = {}
 
+        key_index = {}
+        indexed_data = {}
+        for i, (key, value) in enumerate(data_d.iteritems()) :
+            key_index[i] = key
+            indexed_data[i] = value
+            
         for field in self.blocker.tfidf_fields :
             self.blocker.tfIdfBlock(((record_id, record[field])
                                      for record_id, record 
-                                     in data_d.iteritems()),
+                                     in indexed_data.iteritems()),
                                     field)
 
-        for block_key, record_id in self.blocker(data_d.iteritems()) :
-            blocks.setdefault(str(block_key), []).append((record_id, 
-                                                          data_d[record_id]))
+        for block_key, record_id in self.blocker(indexed_data.iteritems()) :
+            blocks.setdefault(block_key, []).append(record_id) 
+
+        blocks = blocks.values()
+
+        blocks = [records for records in blocks if len(records) > 1]
+
+        blocks = [[(key_index[record_id], indexed_data[record_id])
+                   for record_id in records]
+                  for records in blocks]
 
         # Redundant-free Comparisons from Kolb et al, "Dedoop:
         # Efficient Deduplication with Hadoop"
         # http://dbs.uni-leipzig.de/file/Dedoop.pdf
-        for block_id, (block, records) in enumerate(blocks.iteritems()) :
+        for block_id, records in enumerate(blocks) :
             for record_id, record in records :
                 coverage.setdefault(record_id, []).append(block_id)
 
-        for block_id, (block_key, records) in enumerate(blocks.iteritems()) :
+        blocks = iter(blocks)
+
+        for block_id, records in enumerate(blocks) :
+            if block_id % 10000 == 0 :
+                logger.info("%s blocks" % block_id)
             tuple_records = []
             for record_id, record in records :
                 smaller_ids = set([covered_id for covered_id 
@@ -315,9 +339,11 @@ class RecordLinkMatching(Matching) :
 
     def match(self, data_1, data_2, threshold = 1.5) : # pragma : no cover
         """
-        Identifies pairs of records that refer to the same entity, returns tuples
-        of record ids, where both record_ids within a tuple should refer
-        to the same entity
+        Identifies pairs of records that refer to the same entity, returns
+        tuples containing a set of record ids and a confidence score as a float
+        between 0 and 1. The record_ids within each set should refer to the 
+        same entity and the confidence score is the estimated probability that
+        the records refer to the same entity.
         
         This method should only used for small to moderately sized datasets
         for larger data, use matchBlocks
@@ -363,9 +389,23 @@ class RecordLinkMatching(Matching) :
         blocked_pairs = self._blockData(data_1, data_2)
         return self.thresholdBlocks(blocked_pairs, recall_weight)
 
-    def _blockPairs(self, block) : # pragma : no cover
-        base, target = block
-        return itertools.product(base, target)
+    def _blockedPairs(self, blocks) :
+        """
+        Generate tuples of pairs of records from a block of records
+        
+        Arguments:
+        
+        blocks -- an iterable sequence of blocked records
+        """
+        
+        block, blocks = core.peek(blocks)
+        self._checkBlock(block)
+
+	product = itertools.product
+
+        pairs = (product(base, target) for base, target in blocks)
+
+        return itertools.chain.from_iterable(pairs) 
         
     def _checkBlock(self, block) :
         try :
@@ -402,20 +442,21 @@ class RecordLinkMatching(Matching) :
             self.blocker.tfIdfBlock(fields_1, fields_2, field)
 
 
+        for block_key, record_id in self.blocker(data_2.items()) :
+            blocks.setdefault(block_key, ([], []))[1].append((record_id, 
+                                                              data_2[record_id]))
         for block_key, record_id in self.blocker(data_1.items()) :
-            blocks.setdefault(block_key, ([], []))[0].append((record_id, 
-                                                              data_1[record_id]))
-
-        for block_key, record_id in self.blocker(data_2.iteritems()) :
             if block_key in blocks :
-                blocks[block_key][1].append((record_id, data_2[record_id]))
+                blocks[block_key][0].append((record_id, data_1[record_id]))
 
-        for block_id, (_, sources) in enumerate(blocks.iteritems()) :
+        blocks = blocks.values()
+
+        for block_id, sources in enumerate(blocks) :
             for source in sources :
                 for record_id, record in source :
                     coverage.setdefault(record_id, []).append(block_id)
 
-        for block_id, (block_key, sources) in enumerate(blocks.iteritems()) :
+        for block_id, sources in enumerate(blocks) :
             tuple_block = []
             for source in sources :
                 tuple_source = []
@@ -438,18 +479,19 @@ class StaticMatching(Matching) :
 
     def __init__(self, 
                  settings_file, 
-                 num_processes=1) :
+                 num_cores=None) :
         """
         Initialize from a settings file
         #### Example usage
 
             # initialize from a settings file
-            deduper = dedupe.Dedupe('my_learned_settings')
+            with open('my_learned_settings', 'rb') as f:
+                deduper = dedupe.StaticDedupe(f)
 
         #### Keyword arguments
         
         `settings_file`
-        A file location for a settings file.
+        A file object containing settings data.
 
 
         Settings files are typically generated by saving the settings
@@ -458,21 +500,19 @@ class StaticMatching(Matching) :
         """
         super(StaticMatching, self).__init__()
 
+        if num_cores is None :
+            self.num_cores = multiprocessing.cpu_count()
+        else :
+            self.num_cores = num_cores
 
-        if settings_file.__class__ is not str :
-            raise ValueError("Must supply a settings file name")
-
-        self.num_processes = num_processes
-
-        with open(settings_file, 'rb') as f: # pragma : no cover
-            try:
-                self.data_model = pickle.load(f)
-                self.predicates = pickle.load(f)
-                self.stop_words = pickle.load(f)
-            except KeyError :
-                raise ValueError("The settings file doesn't seem to be in "
-                                 "right format. You may want to delete the "
-                                 "settings file and try again")
+        try:
+            self.data_model = pickle.load(settings_file)
+            self.predicates = pickle.load(settings_file)
+            self.stop_words = pickle.load(settings_file)
+        except KeyError :
+            raise ValueError("The settings file doesn't seem to be in "
+                             "right format. You may want to delete the "
+                             "settings file and try again")
 
         self._logLearnedWeights()
         logger.info(self.predicates)
@@ -492,23 +532,24 @@ class ActiveMatching(Matching) :
     - writeTraining
     - uncertainPairs
     - markPairs
+    - cleanupTraining
     """
 
     def __init__(self, 
-                 field_definition, 
+                 variable_definition, 
                  data_sample = None,
-                 num_processes = 1) :
+                 num_cores = None) :
         """
         Initialize from a data model and data sample.
 
         #### Example usage
 
             # initialize from a defined set of fields
-            fields = {'Site name': {'type': 'String'},
-                      'Address':   {'type': 'String'},
-                      'Zip':       {'type': 'String', 'Has Missing':True},
-                      'Phone':     {'type': 'String', 'Has Missing':True},
-                      }
+            fields = [{'field' : 'Site name', 'type': 'String'},
+                      {'field' : 'Address', 'type': 'String'},
+                      {'field' : 'Zip', 'type': 'String', 'Has Missing':True},
+                      {'field' : 'Phone', 'type': 'String', 'Has Missing':True},
+                     ]
 
             data_sample = [
                            (
@@ -529,26 +570,12 @@ class ActiveMatching(Matching) :
 
         
         #### Additional detail
-        A field definition is a dictionary where the keys are the fields
-        that will be used for training a model and the values are the
-        field specification
 
-        Field types include
+        A field definition is a list of dictionaries where each dictionary
+        describes a variable to use for comparing records. 
 
-        - String
-
-        A 'String' type field must have as its key a name of a field
-        as it appears in the data dictionary and a type declaration
-        ex. `{'Phone': {type: 'String'}}`
-
-        Longer example of a field definition:
-
-
-            fields = {'name':       {'type': 'String'},
-                      'address':    {'type': 'String'},
-                      'city':       {'type': 'String'},
-                      'cuisine':    {'type': 'String'}
-                      }
+        For details about variable types, check the documentation.
+        <http://dedupe.readthedocs.org>`_ 
 
         In the data_sample, each element is a tuple of two
         records. Each record is, in turn, a tuple of the record's key and
@@ -559,11 +586,7 @@ class ActiveMatching(Matching) :
         """
         super(ActiveMatching, self).__init__()
 
-        if field_definition.__class__ is not dict :
-            raise ValueError('Incorrect Input Type: must supply '
-                             'a field definition.')
-
-        self.data_model = DataModel(field_definition)
+        self.data_model = DataModel(variable_definition)
 
         self.data_sample = data_sample
 
@@ -574,8 +597,10 @@ class ActiveMatching(Matching) :
         else :
             self.activeLearner = None
 
-        self.num_processes = num_processes
-
+        if num_cores is None :
+            self.num_cores = multiprocessing.cpu_count()
+        else :
+            self.num_cores = num_cores
 
         training_dtype = [('label', 'S8'), 
                           ('distances', 'f4', 
@@ -586,21 +611,30 @@ class ActiveMatching(Matching) :
                                                            'match': []})
 
 
-
-    def readTraining(self, training_source) : # pragma : no cover
+    def cleanupTraining(self) : # pragma : no cover
         '''
-        Read training from previously saved training data file
+        Clean up data we used for training. Free up memory.
+        '''
+        del self.training_data
+        del self.training_pairs
+        del self.activeLearner
+        del self.data_sample
+
+
+    def readTraining(self, training_file) :
+        '''
+        Read training from previously built training data file object
         
         Arguments:
         
-        training_source -- the path of the training data file
+        training_file -- file object containing the training data
         '''
-
+        
         logger.info('reading training from file')
+        
+        training_pairs = json.load(training_file, 
+                                   cls=serializer.dedupe_decoder)
 
-        with open(training_source, 'r') as f:
-            training_pairs = json.load(f, 
-                                       cls=serializer.dedupe_decoder)
 
         for (label, examples) in training_pairs.items():
             if examples :
@@ -687,32 +721,31 @@ class ActiveMatching(Matching) :
                                      self.stop_words) 
 
 
-    def writeSettings(self, file_name): # pragma : no cover
+    def writeSettings(self, file_obj): # pragma : no cover
         """
-        Write a settings file that contains the 
-        data model and predicates
+        Write a settings file containing the 
+        data model and predicates to a file object
 
         Keyword arguments:
-        file_name -- path to file
+        file_obj -- file object to write settings data into
         """
 
-        with open(file_name, 'w') as f:
-            pickle.dump(self.data_model, f)
-            pickle.dump(self.predicates, f)
-            pickle.dump(self.stop_words, f)
+        pickle.dump(self.data_model, file_obj)
+        pickle.dump(self.predicates, file_obj)
+        pickle.dump(dict(self.stop_words), file_obj)
 
-    def writeTraining(self, file_name): # pragma : no cover
+    def writeTraining(self, file_obj): # pragma : no cover
         """
         Write to a json file that contains labeled examples
 
         Keyword arguments:
-        file_name -- path to a json file
+        file_obj -- file object to write training data to
         """
 
-        with open(file_name, 'wb') as f:
-            json.dump(self.training_pairs, 
-                      f, 
-                      default=serializer._to_json)
+        json.dump(self.training_pairs, 
+                  file_obj, 
+                  default=serializer._to_json,
+                  ensure_ascii=False)
 
 
     def uncertainPairs(self) :
@@ -955,10 +988,56 @@ class RecordLink(RecordLinkMatching, ActiveMatching) :
 
         return data_sample
 
+class GazetteerMatching(RecordLinkMatching) :
+    
+    def __init__(self, *args, **kwargs) :
+        super(GazetteerMatching, self).__init__(*args, **kwargs)
+
+        self._cluster = clustering.gazetteMatching
+        self._linkage_type = "GazetteerMatching"
+
+    def match(self, data_1, data_2, threshold = 1.5, n_matches = 1) : # pragma : no cover
+        """Identifies pairs of records that refer to the same entity, returns
+        tuples containing a set of record ids and a confidence score as a float
+        between 0 and 1. The record_ids within each set should refer to the 
+        same entity and the confidence score is the estimated probability that
+        the records refer to the same entity.
+        
+        This method should only used for small to moderately sized datasets
+        for larger data, use matchBlocks
+        
+        Arguments:
+        data_1    -- Dictionary of records from first dataset, where the 
+                     keys are record_ids and the values are dictionaries
+                     with the keys being field names
+
+        data_2    -- Dictionary of records from second dataset, same form 
+                     as data_1
+                                          
+        threshold -- Number between 0 and 1 (default is .5). We will consider
+                     records as potential duplicates if the predicted 
+                     probability of being a duplicate is above the threshold.
+
+                     Lowering the number will increase recall, raising it
+                     will increase precision
+        
+        n_matches -- Maximum number of possible matches from data_2
+                     for each record in data_1
+        """
+        self._cluster = clustering.gazetteMatching
+        blocked_pairs = self._blockData(data_1, data_2)
+        return self.matchBlocks(blocked_pairs, threshold, n_matches)
+
+class Gazetteer(RecordLink, GazetteerMatching):
+    pass
+
+class StaticGazetteer(StaticRecordLink, GazetteerMatching):
+    pass
 
 def predicateGenerator(data_model) :
-    predicates = []
-    for field, definition in data_model['fields'].items() :
-        predicates.extend(definition.predicates)
+    predicates = set([])
+    for definition in data_model['fields'] :
+        if hasattr(definition, 'predicates') :
+            predicates.update(definition.predicates)
 
     return predicates
